# -*- coding: utf-8 -*-

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import swat_em.analyse
import numpy as np
import pdb


def test_DFT():
    print('Test Harmonic Analysis')
    x = np.linspace(0, 2*np.pi, 11)
    y = 10*np.cos(x) + 3*np.cos(3*x)

    HA = np.abs(swat_em.analyse.DFT(y[:-1]))

    assert 10 == np.round(HA[1], 5)
    assert 3 == np.round(HA[3], 5)





if __name__ == '__main__':
    test_DFT()




# -*- coding: utf-8 -*-

from PyQt5 import uic
from PyQt5.QtWidgets import QDialog, QTableWidgetItem
import sys
import os
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar)
from swat_em.config import config
#  sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__dir__ = os.path.dirname(os.path.abspath(__file__))



class Factors(QDialog):
    def __init__(self, data):
        self.data = data
        super().__init__()
       
        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, 'ui/Factors.ui'), self)
        self.setWindowTitle('Additional factors')
        
        self.groupBox_1.toggled.connect(self.update)
        self.groupBox_2.toggled.connect(self.update)
        self.groupBox_3.toggled.connect(self.update)
        
        self.doubleSpinBox_Ds.valueChanged.connect(self.update)
        self.doubleSpinBox_s.valueChanged.connect(self.update)
        self.doubleSpinBox_skew.valueChanged.connect(self.update)
        self.doubleSpinBox_alpha.valueChanged.connect(self.update)
        

        #  self.radioButton_slayer.toggled.connect(self.change_layers) # disable lineEdit for single phase windings
        
        self.fig = plt.figure(5, dpi=90)
        self.fig.clf()
        self.canvas1 = FigureCanvas(self.fig)
        self.mplvl_factors.addWidget(self.canvas1)
        self.canvas1.draw()
        self.toolbar1 = NavigationToolbar(self.canvas1, 
                self.mplwidget_factors, coordinates=True)
        self.toolbar1.setFixedHeight(25)
        self.mplvl_factors.addWidget(self.toolbar1)
        

    def update(self):
        print('update')


    def run(self):
        self.update()  # calc initial
        ok = self.exec_()
        if ok:
            print('return something')



















# -*- coding: utf-8 -*-
# testcase for individual number of turns

# winding data taken from
# AC-Motoren mit verteilter Zahnspulenwicklung
# Gerhard Huth, Jens Krotsch
# e & i Elektrotechnik und Informationstechnik 2019
# https://dx.doi.org/10.1007/s00502-019-0717-9

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from swat_em.datamodel import datamodel
import swat_em.analyse
import numpy as np


def test_doza():
    Q = 12
    p = 1
    m = 3
    wstep = 1
    layers = 2
    U = [[1,2,-11,-12],[5,6,-7,-8]]
    V = [[-3,-4,5,6],[9,10,-11,-12]]
    W = [[-7,-8,9,10],[1,2,-3,-4]]
    S = [U, V, W]

    zs = 1
    dz =  [ zs*0.0,  zs*0.1,  zs*0.2,  zs*0.3]
    kw1 = [ 0.4830,  0.5054,  0.5278,  0.5502]
    kw5 = [-0.1294, -0.0458, -0.0379, -0.1215]
    kw7 = [ 0.1294,  0.0458,  0.0379,  0.1215]

    for k in range(len(dz)):
        za = zs + dz[k]
        zb = zs - dz[k]
        turns = [[[zb,za,za,zb], [za,zb,zb,za]],
                 [[za,zb,zb,za], [za,zb,zb,za]],
                 [[za,zb,zb,za], [za,zb,zb,za]]]

        data = datamodel()
        data.set_machinedata(Q = Q, p = p, m = m)
        data.set_phases(S, turns = turns)
        data.analyse_wdg()
        #  data.save_to_file('doza_{}.wdg'.format(dz[k]))
        bc, txt = data.get_basic_characteristics()
        #  print(bc)

        idx = data.results['nu_el'].index(1)
        assert kw1[k] == np.round(data.results['kw_el'][idx][0], 4) # phase U
        assert kw1[k] == np.round(data.results['kw_el'][idx][1], 4) # phase V
        assert kw1[k] == np.round(data.results['kw_el'][idx][2], 4) # phase W
        idx = data.results['nu_el'].index(5)
        assert kw5[k] == np.round(data.results['kw_el'][idx][0], 4) # phase U
        assert kw5[k] == np.round(data.results['kw_el'][idx][1], 4) # phase V
        assert kw5[k] == np.round(data.results['kw_el'][idx][2], 4) # phase W
        idx = data.results['nu_el'].index(7)
        assert kw7[k] == np.round(data.results['kw_el'][idx][0], 4) # phase U
        assert kw7[k] == np.round(data.results['kw_el'][idx][1], 4) # phase V
        assert kw7[k] == np.round(data.results['kw_el'][idx][2], 4) # phase W


if __name__ == '__main__':
    test_doza()





# -*- coding: utf-8 -*-
'''
Provides a central class for data 
'''
import json
import os
import fractions
import math
import copy
import string
import gzip
import gc
from swat_em import analyse
from swat_em import report as rep
from swat_em import wdggenerator
from swat_em.config import config




class datamodel:
    '''
    Provides a central place for all data. All analysis functions are
    connect with this class.
    ''' 
    file_format = 1
    machinedata_keys = ['Q', 'p', 'm', 'phases', 'wstep']
    results_keys = ['q', 'nu_el', 'Ei_el', 'kw_el', 'phaseangle_el', 
                    'nu_mech', 'Ei_mech', 'kw_mech', 'phaseangle_mech',
                    'valid', 'error']
    
    def __init__(self):
        self.reset_data()
        self.reset_results()
        #  self.filename = None
        self.actual_state_saved = False
        
    #  def set_filename(self, filename):
        #  self.filename = filename
    
    def reset_data(self):
        '''
        resets all data of the datamodel
        '''
        self.title = ''
        self.notes = ''
        self.machinedata = {}
        for key in self.machinedata_keys:
            self.machinedata[key] = None
        self.machinedata['turns'] = 1
        self.actual_state_saved = False
    
    def reset_results(self):
        self.results = {}
        for key in self.results_keys:
            self.results[key] = None
        self.actual_state_saved = False
    
    def set_title(self, title):
        self.title = title
    
    def get_title(self):
        return self.title
    
    def get_notes(self):
        return self.notes
        
    def set_notes(self, notes):
        self.notes = notes
    
    #  def set_config(self, config):
        #  self.config = config
    
    def set_machinedata(self, Q = None, p = None, m = None):
        '''
        setting the machine data 

        Parameters
        ----------
        Q :      integer
                 number of slots
        p :      integer
                 number of pole pairs
        m:       integer
                 number of phases
        '''
        if Q:
            self.machinedata['Q'] = int(Q)
        if p:
            self.machinedata['p'] = int(p)
        if m:
            self.machinedata['m'] = int(m)
        self.actual_state_saved = False
        
        
    def set_phases(self, S, turns = 1, wstep = None):
        '''
        setting the winding layout 

        Parameters
        ----------
        S : list of lists
            winding layout for every phase, for example:
            S = [[1,-2], [3,-4], [5,-6]]. This means there are 3 phases
            with phase 1 in in slot 1 and in slot 2 with negativ winding
            direction.
        wstep : winding step (slots as unit)
        ''' 
        self.machinedata['phases'] = S
        self.machinedata['turns'] = turns
        self.actual_state_saved = False
        self.set_machinedata(m = len(S))
        self.machinedata['phasenames'] = [string.ascii_uppercase[k] for k in range(len(S))]
        if wstep:
            self.machinedata['wstep'] = int(wstep)
        

    def set_valid(self, valid, error):
        self.results['valid'] = valid
        self.results['error'] = error
    
    
    def genwdg(self, Q, P, m, w, layers, turns = 1):
        '''
        Generates a winding layout and stores it in the datamodel

        Parameters
        ----------
        Q :      integer
                 number of slots
        P :      integer
                 number of poles
        m :      integer
                 number of phases
        w :      integer
                 winding step (1 for tooth coils)
        layers : integer
                 number of coil sides per slot    
        turns  : integer
                 number of turns per coil

        '''
        
        self.set_machinedata(Q, int(P/2), m)
        wdglayout = wdggenerator.genwdg(Q, P, m, w, layers)
        
        self.set_phases(S = wdglayout['phases'], turns = turns, wstep = wdglayout['wstep'])
        self.set_valid(valid = wdglayout['valid'], error = wdglayout['error'])
        self.analyse_wdg()
        self.actual_state_saved = True # Simulate save state


    def get_num_layers(self):
        '''
        Returns the number of layers of the actual winding layout
        '''
        l = 1
        for p in self.machinedata['phases']:
            if len(p[1]) > 1:
                l = 2
        return l

        
    def get_basic_characteristics(self):
        '''
        Returns the basic charactericits of the winding as 
        dictionary and a string
        '''
        if not 'basic_char' in self.results.keys():
            bc = analyse.get_basic_characteristics(
                self.machinedata['Q'],
                2*self.machinedata['p'],
                self.machinedata['m'],
                self.machinedata['phases'],
                self.machinedata['turns'])
            self.results['basic_char'] = bc
        else:
            bc = self.results['basic_char']

        dat = [['Number of slots ',    rep.italic('Q: '),  str(self.machinedata['Q'])],
               ['Number of poles ',    rep.italic('2p: '), str(2*self.machinedata['p'])],
               ['Number of phases ',   rep.italic('m: '),  str(self.machinedata['m'])],
               ['slots per 2p per m ', rep.italic('q: '),  str(bc['q'])]]
        for i, k in enumerate(bc['kw1']):
            dat.append(['winding factor (m={}) '.format(i+1), rep.italic('kw1: '), str(round(k,3)) ])
        dat.append(['lcm(Q, P) ', '', str(bc['lcmQP'])])
        dat.append(['periodic base winding ', rep.italic('t: '), str(bc['t'])])
        dat.append(['parallel connection ', rep.italic('a: '), str(bc['a'])])
        if bc['sym'] and bc['a']:
            dat.append(['symmetric ', '', str(bc['sym'])])
        else:
            dat.append([rep.red('symmetric '), '', rep.red(str(bc['sym']))])
        txt = rep.table(dat)
        if self.results['error']:
            txt.append(rep.red('error: ' + str(self.results['error'])))
        
        if len(self.notes) > 0:
            #  txt.append('<br><br>' + rep.bold(Notes) + ':<br>')
            txt.append(rep.bold('<br><br>Notes<br>'))
            txt.append(rep.italic(self.notes))
        
        txt = ''.join(txt)
        return bc, txt
        
    def calc_q(self):
        self.results['q'] = analyse.calc_q(self.machinedata['Q'], 
                                self.machinedata['p'],
                                self.machinedata['m'])
        
        
    def analyse_wdg(self):
        '''
        analyses the winding (winding factor etc.)
        ''' 
        self.reset_results()
        
        # number of slots per pole and phase
        self.q = fractions.Fraction(self.machinedata['Q'] / (
            self.machinedata['m']*2*self.machinedata['p'])).limit_denominator(100)
        self.results['q'] = self.q
        self.n = self.q.denominator
        self.z = self.q.numerator
        
        # base winding
        self.results['t'] = math.gcd(self.machinedata['Q'], self.machinedata['p'])
        
        # electrical winding factor
        a, b, c, d = analyse.calc_kw(
            self.machinedata['Q'], 
            self.machinedata['phases'], 
            self.machinedata['turns'],
            self.machinedata['p'], 
            config['N_nu_el'],
            config )
        self.results['nu_el'] = a
        self.results['Ei_el'] = b
        self.results['kw_el'] = c
        self.results['phaseangle_el'] = d
        
        # mechanical winding factor
        a, b, c, d = analyse.calc_kw(
            self.machinedata['Q'], 
            self.machinedata['phases'],
            self.machinedata['turns'], 
            1.0, 
            config['N_nu_mech'],
            config )
        self.results['nu_mech'] = a
        self.results['Ei_mech'] = b
        self.results['kw_mech'] = c
        self.results['phaseangle_mech'] = d
        
        # winding symmetric?
        self.results['wdg_is_symmetric'] = analyse.wdg_is_symmetric(self.results['Ei_el'],
            self.machinedata['m'])
        #  print(self.results['wdg_is_symmetric'])
        
        # periodicity of the winding (radial force, parallel connection)? 
        self.results['wdg_periodic'] = analyse.wdg_get_periodic(self.results['Ei_el'], self.machinedata['phases'])
        #  print(self.results['wdg_periodic'])
        self.actual_state_saved = False
        
        # MMK
        phi, MMK, theta = analyse.calc_MMK(self.machinedata['Q'],
                                           self.machinedata['m'],
                                           self.machinedata['phases'],
                                           self.machinedata['turns'])
        nu = list(range(config['max_nu_MMK']+1))
        HA = analyse.DFT(MMK[:-1])[:config['max_nu_MMK']+1]
        self.results['MMK'] = {}
        self.results['MMK']['MMK'] = MMK
        self.results['MMK']['phi'] = phi
        self.results['MMK']['theta'] = theta
        self.results['MMK']['nu'] = nu
        self.results['MMK']['HA'] = HA.tolist()
        
        bc, bc_txt = self.get_basic_characteristics()
        self.results['basic_char'] = bc

        

    def save_to_file(self, fname):
        '''
        Saves the data to file. 

        Parameters
        ----------
        fname :  string
                 file name
        ''' 
        def complex2tuple_1(cplx):
            '''
            converts complex data to tuple of (real, imag)-part
            
            Parameters
            ----------
            fname : list
                    with complex numbers
            '''
            for k1 in range(len(cplx)):
                        cplx[k1] = (cplx[k1].real, cplx[k1].imag)
            return cplx
       
        def complex2tuple_3(cplx):
            '''
            converts complex data to tuple of (real, imag)-part
            
            Parameters
            ----------
            fname : list(list(list))
                    3 nested lists with complex numbers
            '''
            for k1 in range(len(cplx)):
                for k2 in range(len(cplx[k1])):
                    for k3 in range(len(cplx[k1][k2])):
                        cplx[k1][k2][k3] = (cplx[k1][k2][k3].real, cplx[k1][k2][k3].imag)
            return cplx
        
        M = {}
        M['file_format'] = self.file_format
        M['machinedata'] = self.machinedata
        M['results'] = copy.deepcopy(self.results)
        
        # convert lists with complex numbers to tuple with real and imag
        if 'Ei_el' in M['results'].keys():
            M['results']['Ei_el'] = complex2tuple_3(M['results']['Ei_el'])
        
        if 'Ei_mech' in M['results'].keys():
            M['results']['Ei_mech'] = complex2tuple_3(M['results']['Ei_mech'])
        
        if 'MMK' in M['results'].keys():
            M['results']['MMK']['HA'] = complex2tuple_1(M['results']['MMK']['HA'])
        
        # convert fractional number
        if 'q' in M['results'].keys():
            M['results']['q'] = str(M['results']['q'])
            
        if 'basic_char' in M['results'].keys():
            if 'q' in M['results']['basic_char'].keys():
                M['results']['basic_char']['q'] = str(M['results']['basic_char']['q'])
        
        # save as compressed file
        with gzip.GzipFile(fname, 'w') as f:
            s = json.dumps(M, indent = 2)                  # string
            f.write(s.encode('utf-8'))           # bytes        
            
        # save as ASCII file
        #  with open(fname, 'w') as f:
            #  json.dump(M, f, indent = 2)
        self.actual_state_saved = True


    def load_from_file(self, fname):
        '''
        Load data from file. 

        Parameters
        ----------
        fname :  string
                 file name
        ''' 
        
        def tuple2complex_1(cplx):
            '''
            converts tuple of (real, imag) to complex number
            
            Parameters
            ----------
            fname : list
                    with complex numbers
            '''
            for k1 in range(len(cplx)):
                        cplx[k1] = cplx[k1][0] + 1j*cplx[k1][1]
            return cplx
        
        def tuple2complex_3(cplx):
            '''
            converts tuple of (real, imag) to complex number
            
            Parameters
            ----------
            fname : list(list(list))
                    3 nested lists with tuples
            '''
            for k1 in range(len(cplx)):
                for k2 in range(len(cplx[k1])):
                    for k3 in range(len(cplx[k1][k2])):
                        cplx[k1][k2][k3] = cplx[k1][k2][k3][0] + 1j*cplx[k1][k2][k3][1]
            return cplx
        
        
        if os.path.isfile(fname):
            self.reset_results()
            # load ASCII file
            #  with open(fname) as f:
                #  M = json.load(f)
            
            # load compressed file
            with gzip.GzipFile(fname) as f:  # gzip
                s = f.read()                    # bytes
                s = s.decode('utf-8')           # string
                M = json.loads(s)               # data  
            
            self.machinedata = M['machinedata']
            

            
            # convert lists with tuple of real and imag to complex number
            if 'Ei_el' in M['results'].keys():
                M['results']['Ei_el'] = tuple2complex_3(M['results']['Ei_el'])
            
            if 'Ei_mech' in M['results'].keys():
                M['results']['Ei_mech'] = tuple2complex_3(M['results']['Ei_mech'])
            
            if 'MMK' in M['results'].keys():
                M['results']['MMK']['HA'] = tuple2complex_1(M['results']['MMK']['HA'])
            
            # convert fractional number
            if 'q' in M['results'].keys():
                M['results']['q'] = fractions.Fraction(M['results']['q'])
            
            if 'basic_char' in M['results'].keys():
                if 'q' in M['results']['basic_char'].keys():
                    M['results']['basic_char']['q'] = fractions.Fraction(M['results']['basic_char']['q'])
            
                self.results = copy.deepcopy(M['results'])
            self.actual_state_saved = True



class project:
    '''
    Provides all data-objects (all winding in workspace)
    '''
    file_format = 2
    def __init__(self):
        self.models = []
        self.filename = None
        self.undo_state = []
        self.redo_state = []
        
    def save_undo_state(self):
        '''saves the actual state of the models for undo function'''
        self.undo_state.append(copy.deepcopy(self.models))
    
    def save_redo_state(self):
        '''saves the actual state of the models for redo function'''
        self.redo_state.append(copy.deepcopy(self.models))
    
    def reset_undo_state(self):
        '''delete all existings undo state saves - no undo possible any more'''
        self.undo_state = []
    
    def reset_redo_state(self):
        '''delete all existings redo state saves - no redo possible any more'''
        self.redo_state = []
    
    def get_num_undo_state(self):
        '''return the number of undo state saves = number of possible undo opserations'''
        return len(self.undo_state)
    
    def get_num_redo_state(self):
        '''return the number of redo state saves = number of possible redo opserations'''
        return len(self.redo_state)
    
    def undo(self):
        '''restores the last state'''
        if self.get_num_undo_state() > 0:
            self.models = self.undo_state.pop(-1)
            gc.collect()  # force to free memory
    
    def redo(self):
        '''restores the last state'''
        if self.get_num_redo_state() > 0:
            self.models = self.redo_state.pop(-1)
            gc.collect()  # force to free memory
    
    def set_filename(self, filename):
        '''saves the filename if a the data is load from file or
        stored to a file'''
        self.filename = filename
    
    def gen_model_name(self):
        '''
        creates an unique title for a model
        '''
        titles = [data.title for data in self.models]
        i = 0
        while True:
            name = 'untitled' + str(i)
            if name not in titles:
                return name
            i += 1
    
    def add_model(self, data):
        '''
        adds 'data' model to project. Generates a unique title if there
        is no title in 'data'
        '''
        if data.title == '':
            name = self.gen_model_name()
            data.set_title(name)
            data.actual_state_saved = False
        self.models.append(data)
        
    def get_titles(self):
        '''returns the title of all models'''
        return [data.title for data in self.models]

    def get_model_by_index(self, idx):
        '''returns the model with the index 'idx' '''
        return self.models[idx]

    def delete_model_by_index(self, idx):
        '''deletes the model with the index 'idx' '''
        del self.models[idx]
    
    def clone_by_index(self, idx):
        '''duplicates the model with the index 'idx' '''
        data = copy.deepcopy(self.models[idx])
        data.set_title(data.get_title() + '_copy')
        data.actual_state_saved = False
        self.add_model(data)

    def rename_by_index(self, idx, newname):
        '''change the title of the model with the index 'idx' 
        with name 'newname' '''
        self.models[idx].set_title(newname)
    
    def replace_model_by_index(self, newmodel, idx):
        '''replaces the model of index 'idx' with 'newmodel' '''
        self.models[idx] = newmodel

    def set_actual_state_saved(self):
        '''marks all models as saved'''
        for data in self.models:
            data.actual_state_saved = True

    def get_actual_state_saved(self):
        '''returns True if all data is saved'''
        for data in self.models:
            if not data.actual_state_saved:
                return False
        return True


    def save_to_file(self, fname):
        '''
        Saves the data to file. 

        Parameters
        ----------
        fname :  string
                 file name
        ''' 
        M = {}
        M['file_format'] = self.file_format
        M['models'] = []
        for data in self.models:
            N = {}
            N['machinedata'] = data.machinedata
            N['title'] = data.title
            N['notes'] = data.notes
            M['models'].append(N)
            
        # save as ASCII file
        with open(fname, 'w') as f:
            json.dump(M, f, indent = 2)
        self.actual_state_saved = True



    def load_from_file(self, fname):
        '''
        Load data from file. 

        Parameters
        ----------
        fname :  string
                 file name
        ''' 
        try:
            if os.path.isfile(fname):
                with open(fname) as f:
                    M = json.load(f)
        except:
            # file_format 1 is saved compressed
            import gzip
            with gzip.GzipFile(fname) as f:     # gzip
                s = f.read()                    # bytes
                s = s.decode('utf-8')           # string
                M = json.loads(s)               # data  
        
        if M['file_format'] == 1:
            self.models = []
            data = datamodel()
            data.machinedata = M['machinedata']
            data.machinedata['turns'] = 1
            data.analyse_wdg()
            self.add_model(data)
            self.set_actual_state_saved()
        
        elif M['file_format'] == 2:
            self.models = []
            for m in M['models']:
                data = datamodel()
                data.machinedata = m['machinedata']
                data.set_title(m['title'])
                data.set_notes(m['notes'])
                data.analyse_wdg()
                self.add_model(data)
            self.set_actual_state_saved()
        self.set_filename(fname)





# -*- coding: utf-8 -*-
import sys
import os
from PyQt5 import uic
from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog,\
                            QInputDialog, QMessageBox, QListWidgetItem,\
                            QMenu, QAction, QSplashScreen
from PyQt5.QtGui import QIntValidator, QDoubleValidator, QIcon,QPixmap
from PyQt5 import QtCore

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__dir__ = os.path.dirname(os.path.abspath(__file__))

# Create a splash screen while loading the librarys because this
# takes some time
app = QApplication([])  # temp. app needed for splash
splash_pix = QPixmap(os.path.join(__dir__, 'ui/bitmaps/splash.png'))
splash = QSplashScreen(splash_pix, QtCore.Qt.WindowStaysOnTopHint)
splash.show()

import time
import argparse
from swat_em import dialog_genwdg
from swat_em import dialog_about
from swat_em import dialog_notes
from swat_em import dialog_winding_layout
from swat_em import dialog_factors
from swat_em import dialog_settings
from swat_em.config import config, save_config
from swat_em import datamodel
from swat_em import wdggenerator
from swat_em import plots

MSG_TIME = 3000   # time in which the message is displayed in the statusbar



class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.data = datamodel.datamodel()
        self.project = datamodel.project()

        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, 'ui/MainWindow.ui'), self)
        self.setWindowTitle('SWAT-EM')

        self.DIALOG_GenWinding = dialog_genwdg.GenWinding2()
        self.DIALOG_GenWindingCombinations = dialog_genwdg.GenWindingCombinations()
        self.DIALOG_AdditionalFactors = dialog_factors.Factors(self.data)

        # Connect the buttons
        self.Button_exit.clicked.connect(self.close)
        
        self.projectlist_Button_new.clicked.connect(self.dialog_new_winding)
        self.projectlist_Button_delete.clicked.connect(self.projectlist_delete)
        self.projectlist_Button_clone.clicked.connect(self.projectlist_clone)
        self.projectlist_Button_notes.clicked.connect(self.dialog_get_notes)
        self.projectlist_Button_manual.clicked.connect(self.dialog_EditWindingLayout)
        self.projectlist_Button_auto.clicked.connect(self.dialog_GenWinding)
        self.projectlist_Button_table.clicked.connect(self.dialog_GenWindingCombinations)
        
        # Connect menu
        self.actionExit.triggered.connect(self.close)
        self.actionGenerate_winding.triggered.connect(self.dialog_GenWinding)
        self.actionGenerate_winding_combinations.triggered.connect(self.dialog_GenWindingCombinations)
        
        self.actionsave.triggered.connect(self.save_to_file)
        self.actionsave_as.triggered.connect(self.save_as_to_file)
        self.actionload.triggered.connect(self.load_from_file)
        self.actionProgram_Info.triggered.connect(dialog_about.about)
        self.actionShow_winding_layout.triggered.connect(self.dialog_EditWindingLayout)
        self.actionAdditional_factors.triggered.connect(self.dialog_AdditionalFactors)
        self.actionSettings.triggered.connect(self.dialog_settings)
        self.actionundo.triggered.connect(self.undo)
        self.actionredo.triggered.connect(self.redo)
        
        # Connect project models
        self.project_listWidget.currentRowChanged.connect(self.update_project) 
        self.project_listWidget.itemChanged.connect(self.projectlist_rename)    # item renamed


                        
        # context-menu on project models
        self.project_listWidget.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)

        self.actionClone = QAction("clone", self.project_listWidget, 
                   icon=QIcon(os.path.join(__dir__,'ui/icons/clone.png')))
        self.project_listWidget.addAction(self.actionClone)
        self.actionClone.triggered.connect(self.projectlist_clone)
        
        self.actionDelete = QAction("delete", self.project_listWidget, 
                   icon=QIcon(os.path.join(__dir__,'ui/icons/delete.png')))        
        self.project_listWidget.addAction(self.actionDelete)
        self.actionDelete.triggered.connect(self.projectlist_delete)



        # TODO
        self.actionprint_report.triggered.connect(lambda: QMessageBox.information(self, 'Information', 'Not implemented yet'))
        self.actionHelp.triggered.connect(lambda: QMessageBox.information(self, 'Information', 'Not implemented yet'))
        
        self.plot_tabs.currentChanged.connect(lambda: self.update_plot_in_GUI(small_update = False))
        self.radioButton_electrical.toggled.connect(self.update_plot_in_GUI)
        self.comboBox_star_harmonics.currentIndexChanged.connect(self.update_plot_in_GUI)
        self.checkBoxForceX.toggled.connect(self.update_plot_in_GUI)
        

        #  initial windings  -----------------------------------
        self.data = datamodel.datamodel()
        self.data.genwdg(Q = 12, P = 10, m = 3, w = 1, layers = 2)
        self.data.set_notes('Winding example for a tooth-coil-winding for 12 slots, 10 poles machine')
        self.project.add_model(self.data)

        self.data = datamodel.datamodel()
        self.data.genwdg(Q = 18, P = 2, m = 3, w = 7, layers = 2)
        self.project.add_model(self.data)
        
        self.data = datamodel.datamodel()
        self.data.genwdg(Q = 6, P = 4, m = 3, w = 1, layers = 2, turns = 10)
        self.project.add_model(self.data)
        self.project.set_actual_state_saved()  # only for the initial winding
        
        
        # Plots
        self.MMK_phase_edit.setValidator(QDoubleValidator(0, 360, 1))
        self.MMK_phase_edit.setText(str(0.000))
        
        #  self.MMK_phase_slider.valueChanged.connect(self.update_plot_in_GUI)
        self.MMK_phase_slider.valueChanged.connect(self.update_MMK_phase_edit)
        self.MMK_phase_edit.textEdited.connect(self.update_MMK_phase_slider)
        self.MMK_phase_edit.textChanged.connect(lambda: self.update_plot_in_GUI(small_update = True))
        
        self.fig1 = plots.slot_plot(self.mplvl_slot, self.mplwidget_slot, self.data)
        self.fig2 = plots.slot_star(self.mplvl_star, self.mplwidget_star, self.data, self.tableWidget_star)
        self.fig3 = plots.windingfactor(self.mplvl_wf, self.mplwidget_wf, self.data, self.tableWidget_wf)
        self.fig4 = plots.mmk(self.mplvl_mmk, self.mplwidget_mmk, self.data, self.tableWidget_mmk)
        
        self.update_project_list()
        #  self.update_data_in_GUI()     # not neccessary because of 'update_project_list()
        self.project.reset_undo_state()  # init-winding shouldn't undoable
        self.actionundo.setDisabled(True)
        self.actionredo.setDisabled(True)
        self.show()
    
    
    def save_undo_state(self):
        '''save the actual state of the project models for undo operation'''
        self.project.save_undo_state()
        self.actionundo.setDisabled(False)
        self.project.reset_redo_state() # redo states aren't valid any more
        self.actionredo.setDisabled(True)
        
    
    def undo(self):
        '''undo the last operation (restore state)'''
        self.project.save_redo_state()
        self.actionredo.setDisabled(False)
        self.project.undo()
        self.update_project_list()
        #  self.update_project()
        if self.project.get_num_undo_state() < 1:
            self.actionundo.setDisabled(True) # no more states to restore
        
    def redo(self):
        self.project.save_undo_state()
        self.actionundo.setDisabled(False)
        self.project.redo()
        self.update_project_list()
        #  self.update_project()
        if self.project.get_num_redo_state() < 1:
            self.actionredo.setDisabled(True) # no more states to restore
    
    
    def update_project_list(self, switch_to_new = False):
        '''fill the list of all data objects
        if 'switch_to_new = True: mark the last (new) object in project list 
        and show the content of it. If False: The last idx is marked again'''
        self.project_listWidget.blockSignals(True)
        idx = self.project_listWidget.currentRow() # save current idx
        self.project_listWidget.clear()            # clear existing data
        for i, item in enumerate(self.project.get_titles()):
            widgetitem = QListWidgetItem(item)
            widgetitem.setFlags(widgetitem.flags() | QtCore.Qt.ItemIsEditable)
            self.project_listWidget.addItem(widgetitem)
        self.project_listWidget.blockSignals(False)
        if switch_to_new:
            idx = i
        else:
            idx = i if i < idx else idx
            idx = 0 if idx < 0 else idx
        self.project_listWidget.setCurrentRow(idx)   # last item is current model
    
    def update_project(self, idx = None):
        '''update the clicked data object form the project'''
        if not idx:
            idx = self.project_listWidget.currentRow()
        self.data = self.project.get_model_by_index(idx)
        self.update_data_in_GUI()  # update all data output and figures
    
    def projectlist_delete(self):
        if self.project_listWidget.count() > 1:
            self.save_undo_state()  # for undo
            idx = self.project_listWidget.currentRow()
            self.project.delete_model_by_index(idx)
            self.statusbar.showMessage('winding deleted', MSG_TIME)
            self.update_project_list()
            self.statusbar.showMessage('winding cloned', MSG_TIME)
    
    def projectlist_clone(self):
        self.save_undo_state()  # for undo
        idx = self.project_listWidget.currentRow()
        self.project.clone_by_index(idx)
        self.statusbar.showMessage('winding cloned', MSG_TIME)
        self.update_project_list(switch_to_new = True)

    def projectlist_rename(self):
        self.save_undo_state()  # for undo
        idx = self.project_listWidget.currentRow()
        newname = self.project_listWidget.item(idx).text()
        self.project.rename_by_index(idx, newname)
    
    
    def update_MMK_phase_edit(self):
        '''slider for MMK phase moved'''
        self.MMK_phase_edit.setText(str(self.MMK_phase_slider.value()))


    def update_MMK_phase_slider(self):
        '''lineedit for MMK phase changed'''
        self.MMK_phase_slider.setValue(float(self.MMK_phase_edit.text()))


    def dialog_GenWinding(self):
        '''
        calls the dialog to generate a single winding
        '''
        ret = self.DIALOG_GenWinding.run()
        if ret:
            self.save_undo_state()
            wdglayout = wdggenerator.genwdg(ret['Q'], ret['P'], ret['m'], ret['w'], ret['layers'])
            
            data = datamodel.datamodel()
            data.set_machinedata(Q = ret['Q'], p = ret['P']//2, m = ret['m'])
            data.set_phases(wdglayout['phases'], wstep = wdglayout['wstep'])            
            data.analyse_wdg()
            data.set_valid(valid = wdglayout['valid'], error = wdglayout['error'])            
            if ret['overwrite']:
                self.data = data
                self.project.replace_model_by_index(data, self.project_listWidget.currentRow())
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new = True)
            self.update_data_in_GUI()


    def dialog_GenWindingCombinations(self):
        '''
        calls the dialog to generate a winding based on a big table of available windings
        '''
        ret = self.DIALOG_GenWindingCombinations.run()
        if ret:
            self.save_undo_state()
            wdglayout = wdggenerator.genwdg(ret['Q'], ret['P'], ret['m'], ret['w'], ret['layers'])
            
            data = datamodel.datamodel()
            data.set_machinedata(Q = ret['Q'], p = ret['P']//2, m = ret['m'])
            data.set_phases(wdglayout['phases'], wstep = wdglayout['wstep'])            
            data.analyse_wdg()
            data.set_valid(valid = wdglayout['valid'], error = wdglayout['error'])            
            if ret['overwrite']:
                self.data = data
                self.project.replace_model_by_index(data, self.project_listWidget.currentRow())
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new = True)
            self.update_data_in_GUI()


    def dialog_EditWindingLayout(self):
        '''
        calls the dialog to generate a winding based on a big table of available windings
        '''
        DIALOG_EditWindingLayout = dialog_winding_layout.layout(self.data)
        ret = DIALOG_EditWindingLayout.run()

        if ret:
            self.save_undo_state()
            
            data = datamodel.datamodel()
            data.set_machinedata(Q = ret['Q'], p = ret['P']//2, m = ret['m'])
            data.set_phases(ret['phases'], wstep = ret['w'])            
            data.analyse_wdg()
            #  data.set_valid(valid = wdglayout['valid'], error = wdglayout['error'])
            if ret['overwrite']:
                self.data = data 
                self.project.replace_model_by_index(data, self.project_listWidget.currentRow())
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new = True)
            self.update_data_in_GUI()


    def dialog_AdditionalFactors(self):
        ret = self.DIALOG_AdditionalFactors.run()


    def dialog_get_notes(self):
        '''dialog asks user to type notes and stores it to datamodel'''
        dialog = dialog_notes.get_notes(self.data.notes)
        text = dialog.run()
        if text is not None:
            self.save_undo_state()
            self.data.set_notes(text)
            self.update_data_in_GUI()
        
    def dialog_new_winding(self):
        '''
        dialog to generate a new winding
        wraps the user input to the existing generators
        '''
        dialog = dialog_genwdg.NewWinding()
        ret = dialog.run()
        if ret is not None:
            if ret == 0:   # trigger the different generators
                self.dialog_EditWindingLayout()
            elif ret == 1:
                self.dialog_GenWinding()
            elif ret == 2:
                self.dialog_GenWindingCombinations()
            else:
                raise(Exception, 'winding generator {} not implemented'.format(ret))


    def dialog_settings(self):
        global config
        config
        DIALOG_Settings = dialog_settings.Settings(config)
        ret = DIALOG_Settings.run()
        if ret:
            config = ret
            save_config(config)
            self.data.analyse_wdg()   # recalc the winding model
            self.update_data_in_GUI()


    def update_data_in_GUI(self):
        '''
        plots all winding data in gui
        '''
        bc, bc_str = self.data.get_basic_characteristics()
        
        self.textBrowser_wdginfo.setHtml(bc_str)
        self.update_plot_in_GUI()
        
        self.comboBox_star_harmonics.blockSignals(True)   # prevent double plotting on startup
        self.comboBox_star_harmonics.clear()
        self.comboBox_star_harmonics.addItems([str(k) for k in self.data.results['nu_el']])
        self.comboBox_star_harmonics.blockSignals(False)
        
        
    def update_plot_in_GUI(self, small_update = False):
        '''
        update the current figure
        if 'small_update' is True, only new lines/bars are plottet and not the 
        all axes etc. --> speed up, for MMK-phase slider for example
        '''
        # Update Figures
        idx = self.plot_tabs.currentIndex()
        #  t1 = time.time()
        # update only the current tab
        if idx == 0:
            self.fig1.plot_slots(self.data.machinedata['Q'])
            self.fig1.plot_coilsides(self.data)
        if idx == 1: 
            self.fig2.plot_star(self.data, harmonic_idx = self.comboBox_star_harmonics.currentIndex(),
            ForceX = self.checkBoxForceX.isChecked())
        if idx == 2:
            if self.radioButton_electrical.isChecked():
                self.fig3.plot_windingfactor(self.data, mechanical=False)
            elif self.radioButton_mechanical.isChecked():
                self.fig3.plot_windingfactor(self.data, mechanical=True)
        if idx == 3:  
            self.fig4.plot_mmk(self.data, float(self.MMK_phase_edit.text()), small_update = small_update)
        #  print('duration for plot:', time.time()-t1)

        
    def save_to_file(self):
        '''
        returns True if file is saved
        '''
        if not self.project.filename:
            ret = self.save_as_to_file()
            return ret
        else:
            self.project.save_to_file(self.project.filename)
            self.statusbar.showMessage('Project saved as ' + self.project.filename, MSG_TIME)
            return True


    def save_as_to_file(self):
        '''
        returns True if file is saved
        '''
        options = QFileDialog.Options()
        #  options |= QFileDialog.DontUseNativeDialog
        filename, _ = QFileDialog.getSaveFileName(self,"Save winding file", "","Winding Files (*.wdg)", options=options)
        if filename:
            if not filename.endswith('.wdg'):
                filename += '.wdg'
            self.project.set_filename(filename) 
            self.project.save_to_file(self.project.filename)
            self.statusbar.showMessage('Project saved as ' + self.project.filename, MSG_TIME)
            return True
        else:
            return False


    def load_from_file(self):
        if not self.project.get_actual_state_saved():
            ok = QMessageBox.question(self, 'Exit program', "Do you want to save?", QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel, QMessageBox.Cancel)
            if ok == QMessageBox.Yes:
                ret = self.save_to_file()
                if not ret: # user has not saved
                    return
            if ok == QMessageBox.Cancel: # cancel
                return

        
        options = QFileDialog.Options()
        #  options |= QFileDialog.DontUseNativeDialog  # use qt5 dialog instead of os default
        filename, _ = QFileDialog.getOpenFileName(self,"Open winding file", "","Winding Files (*.wdg)", options=options)
        if filename:
            self.project.load_from_file(filename)
            self.statusbar.showMessage('Project loaded: ' + filename, MSG_TIME)
            self.project.set_filename(filename)
            self.update_project_list()
            self.update_data_in_GUI()
        
    
    def closeEvent(self, event):
        # Exit programm
        
        # Ask for saving
        if self.project.get_actual_state_saved():
            event.accept() # let the window close
        else:
            #  print('ask for saving')
            ok = QMessageBox.question(self, 'Exit program', "Do you want to save?", QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel, QMessageBox.Cancel)
            if ok == QMessageBox.Yes:
                ret = self.save_to_file()
                if ret:  # user has saved
                    event.accept() # let the window close
                else:    # user has not saved
                    event.ignore()
            elif ok == QMessageBox.No:
                event.accept() # close the window
            else:
                event.ignore() # let the window open
                


def main():
    # command line options
    parser = argparse.ArgumentParser(
        prog = 'SWAT-EM',
        description='Generating and evaluating of windings for electrial machines', 
        formatter_class=argparse.RawTextHelpFormatter)
    
    parser.add_argument(action='store', nargs='?', default='', dest='arg_1', help='*.py scriptfile or *.wdg file')
    parser.add_argument('--load', action='store', default='', dest='loadfile', help='Open an existing *.wdg file')
    parser.add_argument('--script', action='store', default='', dest='scriptfile', help='Open an existing *.py script file')
    args = parser.parse_args()
    
    if len(args.arg_1) > 0:
        #  print('my arg: ', args.arg_1)
        ext = os.path.splitext(args.arg_1)[-1]
        if ext.upper() == '.PY':
            args.scriptfile = args.arg_1
        elif ext.upper() == '.WDG':
            args.loadfile = args.arg_1
    
    if args.scriptfile:
        print('eval script file:', args.scriptfile)
    else:
        print('running GUI')
        app = QApplication(sys.argv)
        ex = MainWindow()
        
        if args.loadfile:
            if os.path.isfile(args.loadfile):
                ex.data.load_from_file(args.loadfile)
                ex.data.set_filename(args.loadfile)
                ex.update_data_in_GUI()
            else:
                raise(Exception, 'file not found:'.format(args.loadfile))
        splash.close()
        sys.exit(app.exec_())


if __name__ == '__main__':
    main()
    
    
    



# -*- coding: utf-8 -*-

# Functions for creating windings

import fractions
#  from collections import deque
import math
import numpy as np


def is_even(val):
    '''
    Returns True if absolute value is even
    '''
    return True if abs(val) % 2 is 0 else False


def genwdg(Q, P, m, w, layers):
    """
    Generates a winding layout.

    Parameters
    ----------
    Q :      integer
             number of slots
    P :      integer
             number of poles
    m :      integer
             number of phases
    w :      integer
             winding step (1 for tooth coils)
    layers : integer
             number of coil sides per slot      
             
    Returns
    -------
    return : list
             winding layout (right and left layers) for every phase
    """  
    #  q = fractions.Fraction(Q/(m*P)).limit_denominator(100)
    #  n = q.denominator
    #  z = q.numerator
    
    #  wdglayout = None
    #  if q >= 1:    # overlapping winding
        #  if layers == 1:
            #  wdglayout = overlapping_fractional_slot_slayer(Q, P, m)
        #  else:
            #  wdglayout = overlapping_fractional_slot_dlayer(Q, P, m, w)
    #  else:         # tooth coil winding
        #  if layers == 1:
            #  wdglayout = toothcoil_slayer(Q, P, m)
        #  else:
            #  wdglayout = toothcoil_dlayer(Q, P, m)
    #  if not wdglayout:
        #  print('No winding found')
    
    ret = winding_from_star_of_slot(Q, P, m, w, layers)
        
    return ret




'''


def toothcoil_dlayer(Q, P, m):
    """
    Generates winding layout for double layer tooth coil winding 

    Parameters
    ----------
    Q :      integer
             number of slots
    P :      integer
             number of poles
    m :      integer
             number of phases
             
    Returns
    -------
    return : list
             winding layout for every phase
    """ 
    q = fractions.Fraction(Q/(m*P)).limit_denominator(100)
    n = q.denominator
    z = q.numerator
    cpph = int(Q / m)    # coils per phase
    
    phases = {}
    sign = 1
    i = 1
    for k in range(int(cpph/z)): # for every smallest periodic part of the winding
        for km in range(m):
            if km not in phases.keys():
                phases[km] = [[],[]]
            for k in range(z):
                phases[km][0].append(i*sign)
                i += 1
                phases[km][1].append(-i*sign)
                sign *= -1
            sign *= -1

    # The algorithm starts with a half slot: 1, 2, 2, 3, 3 ... Q+1
    # So the last slot is Q+1 and have to be reduced to slot 1
    for key, value in phases.items():
        for i, k in enumerate(value[0]):
            if k > Q:
                phases[key][0][i] = 1
            elif k < -Q:
                phases[key][0][i] = -1
        for i, k in enumerate(value[1]):
            if k > Q:
                phases[key][1][i] = 1
            elif k < -Q:
                phases[key][1][i] = -1
        
    phases_list = [value for key, value in phases.items()]
    return phases_list


def toothcoil_slayer(Q, P, m):
    """
    Generates winding layout for single layer tooth coil winding 

    Parameters
    ----------
    Q :      integer
             number of slots
    P :      integer
             number of poles
    m :      integer
             number of phases
             
    Returns
    -------
    return : list
             winding layout for every phase
    """  
    q = fractions.Fraction(Q/(m*P)).limit_denominator(100)
    n = q.denominator
    z = q.numerator
    cpph = int(Q / m / 2)    # coils per phase
    
    phases = {}
    sign = 1
    i = 1
    if z%2 == 0: # z is even
        for k in range(int(cpph/z*2)): # for every smallest periodic part of the winding
            for km in range(m):
                if km not in phases.keys():
                    phases[km] = []
                for k in range(int(z/2)):
                    phases[km].append(i*sign)
                    i += 1
                    phases[km].append(-i*sign)
                    i += 1
                sign *= -1
                
    else: # z is odd
        toggle = 1
        for k in range(int(cpph/z*2)): # for every smallest periodic part of the winding
            for km in range(m):
                if km not in phases.keys():
                    phases[km] = []
                for k in range(int((z+toggle)/2)):
                    phases[km].append(i*sign)
                    i += 1
                    phases[km].append(-i*sign)
                    i += 1
                sign *= -1                    
                toggle *= -1

    phases_list = [[value] for key, value in phases.items()]
    return phases_list
'''


def overlapping_fractional_slot_slayer(Q, P, m):
    from collections import deque
    p = P/2
    q = fractions.Fraction(Q/(m*2*p)).limit_denominator(100)
    g = 0  # number of full slots
    while q > 1:
        q -= 1
        g += 1
    
    z2 = q.numerator
    n = q.denominator
    
    # build sequence of coil sides (half of the coil side pairs of the coils)
    cgroups = []    
    for k in range(n):
        if k < z2:
            cgroups.append(g+1)
        if k < (n-z2):
            cgroups.append(g)
    cgroups *= m
    
    # build sequnce of coil sides for each phase (complete pair)
    cgroups_per_phase = [[] for k in range(m) ]
    i = 0
    for g in cgroups:        
        cgroups_per_phase[i].append(g)
        cgroups_per_phase[i].append(-g)
        i += 1
        if i >= m:
            i = 0
    
    if m == 3: # change phase sequence for generating the winding layout
        cgroups_per_phase[1], cgroups_per_phase[2] = cgroups_per_phase[2], cgroups_per_phase[1]
    
    for i in range(m):
        if i%2 is not 0:
            l = deque(cgroups_per_phase[i])
            l.rotate(1)
            cgroups_per_phase[i] = list(l)
    
    # print(cgroups_per_phase)
    # print('z2', z2)
    # print('n', n)
    # print(cgroups)
    
    phases = [[] for k in range(m) ]
    i = 1
    while i < Q:
        for cs in zip(*cgroups_per_phase):
            for km in range(len(cs)):
                for s in range(abs(cs[km])):
                    if cs[km] > 0:
                        phases[abs(km)].append(-i)
                    else:
                        phases[abs(km)].append(i)
                    i += 1
    # print('phases', phases)

    if m == 3: # change phase sequence again
        phases[1], phases[2] = phases[2], phases[1]
    
    for k in range(len(phases)):
        phases[k] = [phases[k], []]
    return phases

'''
# Noch zu testen!!!
def overlapping_fractional_slot_dlayer(Q, P, m, w):
    p = P/2
    q = fractions.Fraction(Q/(m*2*p)).limit_denominator(100)
    g = 0  # number of full slots
    while q > g:
        q -= 1
        g += 1    
    z2 = q.numerator
    n = q.denominator
    
    # build sequence of coil sides (half of the coil side pairs of the coils)
    cgroups = []    
    for k in range(n):
        if k < z2:
            cgroups.append(g+1)
        if k < (n-z2):
            cgroups.append(g)
    cgroups *= m
    
    # build sequnce of coil sides for each phase (complete pair)
    cgroups_per_phase = [[] for k in range(m) ]
    i = 0
    for g in cgroups:        
        cgroups_per_phase[i].append(g)
        cgroups_per_phase[i].append(-g)
        i += 1
        if i >= m:
            i = 0
    
    if m == 3: # change phase sequence for generating the winding layout
        cgroups_per_phase[1], cgroups_per_phase[2] = cgroups_per_phase[2], cgroups_per_phase[1]
    
    for i in range(m):
        if i%2 is not 0:
            l = deque(cgroups_per_phase[i])
            l.rotate(1)
            cgroups_per_phase[i] = list(l)    

    phases = [[[], []] for k in range(m) ]
    i = 1
    while i < Q:
        for cs in zip(*cgroups_per_phase):
            for km in range(len(cs)):
                for s in range(abs(cs[km])):
                    if cs[km] > 0:
                        phases[abs(km)][0].append(-i)
                    else:
                        phases[abs(km)][0].append(i)
                    i += 1
    # print('phases', phases)
    # print('w', w)

    if m == 3: # change phase sequence again
        phases[1], phases[2] = phases[2], phases[1]
    
    for km in range(m):
        for ks in range(len(phases[km][0])):
            tmp = phases[km][0][ks]
            tmp2 = abs(tmp) + w
            if tmp2 < 1:
                tmp2 += Q
            if tmp2 > Q:
                tmp2 -= Q            
            if tmp > 0:    # change polarity (opposite coils side)
                tmp2 *= -1
                        
            phases[km][1].append(tmp2)
    
    # print('phases', phases)
    
    
    
    
    
    
    
    return phases

'''


def winding_from_star_of_slot(Q, P, m, w=-1, layers=2):
    #  if layers == 1:
        #  if w == -1:
            #  w = Q // P
            #  if w <= 0:
                #  w = 1
    #  elif layers == 2:
        #  if w == -1:
            #  w = Q // P
            #  if w <= 0:
                #  w = 1
    if w == -1:
        w = Q // P
        if w <= 0:
            w = 1
    # single layer is only feasable if winding step is odd
    # reduce windings step if it is even
    # if layers == 1:
        # if w > 1 and is_even(w):
            # w -= 1
    # print('w', w)
    p = P//2
    
    error = ''
    valid = True
    
    t = math.gcd(Q, p)
    # Test if there are enough slots for the winding
    if layers == 1:
        test1 = fractions.Fraction(Q / (2*m)).limit_denominator(100)
        if test1.denominator is not 1:
            valid = False
            error += 'For single layer winding Q/(2*m) must be an integer\n'
    elif layers == 2:
        test1 = fractions.Fraction(Q / m).limit_denominator(100)
        if test1.denominator is not 1:
            valid = False
            error += 'For double layer winding Q/m must be an integer'
        

    test2 = fractions.Fraction(Q / (m*t)).limit_denominator(100)
    if test2.denominator is not 1:
        valid = False
        error += 'winding not feasible'

    phasors = []
    for k in range(Q):
        a = 2*np.pi*p/Q * k
        a += np.pi/m/4         # shift by half sector range instead of shifting
        a -= 2*np.pi*p/Q/100   # add a small shift because a phasor could match
                               # a sector border exactly
        
        while a > 2*np.pi:     # angle between 0...2pi
            a -= 2*np.pi
        phasors.append(a)

    phases = [[[],[]] for k in range(m)]
    for km in range(m):                   # for every phase
        r = np.pi/m                       # range for sector
        if is_even(m):                    # shifting m=2 (and maybe m=4, 6,...) 
            mp = 1                        # with special treatment
        else:
            mp = 2
        r1 = [mp*km*r, mp*km*r+r]         # sector for positive coils sides
        r2 = [r1[0]+np.pi, r1[1]+np.pi ]  # sector for negative coils sides

        for k in range(len(r1)):          # sector between 0..2pi
            while r1[k] > 2*np.pi:
                r1[k] -= 2*np.pi    
        for k in range(len(r2)):
            while r2[k] > 2*np.pi:
                r2[k] -= 2*np.pi
        
        for i, p in enumerate(phasors):
            # print(i+1, p, r1, r2)
            if p > r1[0] and p <= r1[1]:       # phasor is in positive sector
                phases[km][0].append(i+1)     # slot belongs to the phase
                c = i+1+w
                while c > Q:
                    c -= Q
                phases[km][1].append(-c)       # add the second layer
            if p > r2[0] and p <= r2[1]:       # phasor is in negative sector
                phases[km][0].append(-i-1)            
                c = i+1+w
                while c > Q:
                    c -= Q
                phases[km][1].append(c)


    # remove second layer. Attention: It is important to fill the second
    # layer with the algorithm above, even if single layer is to generate,
    # because there are special cases, where the single layer winding have
    # to be generated from the double layer winding!
    if layers == 1:    # only 
        if not is_even(w):
            for km in range(m):
                a = phases[km]
                a2 = [[], []]
                for k in range(len(a[0])):
                    #  print(a[0][k], a[1][k])
                    if not is_even(a[0][k]):    # use only the odd slots
                        a2[0].append(a[0][k])
                        a2[0].append(a[1][k])   # from second layer to first layer
                phases[km] = a2
        else:
            # use fallback function for w = even!
            q = fractions.Fraction(Q/(m*P)).limit_denominator(100)
            if q.denominator == 1:
                for km in range(m):
                    phases[km][1] = []
            else:
                #  print('single layer winding not suitable', Q, P)
                phases = overlapping_fractional_slot_slayer(Q, P, m)
                w = Q/P

            
    ret = {'phases': phases, 'wstep': w, 'valid': valid, 'error': error }
    return ret



# -*- coding: utf-8 -*-

from PyQt5 import uic
from PyQt5.QtWidgets import QDialog, QMessageBox
import sys
import os
from swat_em import config

__dir__ = os.path.dirname(os.path.abspath(__file__))

class Settings(QDialog):
    def __init__(self, config):
        self.config = config
        super().__init__()
        #  self.resetted = False # flag is set, if user reset the config
       
        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, 'ui/Settings.ui'), self)
        self.setWindowTitle('Settings')
        
        #  self.Button_CANCEL.clicked.connect(self.reject)
        #  self.Button_OK.clicked.connect(self.accept)
        self.Button_RESET.clicked.connect(self.reset)

        # Nothing is done, when a value is changed, yet
        # later here can be checked for plausibility
        
        #  self.spinBox_nu_el.valueChanged.connect(self.update)
        #  self.spinBox_nu_mech.valueChanged.connect(self.update)
        #  self.spinBox_nu_MMK.valueChanged.connect(self.update)
        #  self.doubleSpinBox_threshold_wf.valueChanged.connect(self.update)
        
    def reset(self):
        ok = QMessageBox.question(self, 'Reset settings', "Do you want to reset the settings to default values?", QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if ok == QMessageBox.Yes:
            #  self.resetted = True
            self.config = config.get_init_config()
            self.write_to_gui()

    # Nothing is done, when a value is changed, yet
    # later here can be checked for plausibility
    #  def update(self):
        #  print('update')


    def write_to_gui(self):
        '''
        Write the config values to the cui
        '''
        self.spinBox_nu_el.setValue(self.config['N_nu_el'])
        self.spinBox_nu_mech.setValue(self.config['N_nu_mech'])
        self.spinBox_nu_MMK.setValue(self.config['max_nu_MMK'])
        self.doubleSpinBox_threshold_wf.setValue(self.config['kw_min'])
        
        self.spinBox_DPI.setValue(self.config['plt']['DPI'])
        self.doubleSpinBox_lw.setValue(self.config['plt']['lw'])
        self.doubleSpinBox_lw_thin.setValue(self.config['plt']['lw_thin'])

    def read_from_gui(self):
        '''
        Read the config values from the cui
        '''
        self.config['N_nu_el'] = self.spinBox_nu_el.value()
        self.config['N_nu_mech'] = self.spinBox_nu_mech.value()
        self.config['max_nu_MMK'] = self.spinBox_nu_MMK.value()
        self.config['kw_min'] = self.doubleSpinBox_threshold_wf.value() 
        
        self.config['plt']['DPI'] = self.spinBox_DPI.value()
        self.config['plt']['lw'] = self.doubleSpinBox_lw.value()
        self.config['plt']['lw_thin'] = self.doubleSpinBox_lw_thin.value()
        
        


    def run(self):        
        self.write_to_gui()
        
        ok = self.exec_()
        if ok:
            self.read_from_gui()
            return self.config



















# -*- coding: utf-8 -*-
'''
Provides functions for reports
''' 

def italic(txt):
    return '<i>' + txt + '</i>'


def table_header(data):
    txt = ['<tr>']
    for d in data:
        txt.append('<th>')
        txt.append(d)
        txt.append('</th>')
    txt.append('</tr>')
    return ''.join(txt)


def table_row(data, col = None):
    if col:
        txt = ['<tr bgcolor="{}">'.format(col)]
    else:
        txt = ['<tr>']
    for d in data:
        txt.append('<td>')
        txt.append(d)
        txt.append('</td>')
    txt.append('</tr>')
    return ''.join(txt)


def red(text):
    txt = ['<span style=\" color:#ff0000;\" >']
    txt.append(text)
    txt.append('</span>')
    return ''.join(txt)

def italic(text):
    return '<i>' + text + '</i>'

def bold(text):
    return '<b>' + text + '</b>'

def table(dat):
    txt = ['<table style="width:100%">']
    col1 = '#FFFFFF'
    col2 = '#DDDADA'
    for i, d in enumerate(dat):
        col = col1 if i%2 != 0 else col2
        txt.append(table_row(d, col=col))
    
    txt.append('</table>')
    return txt



# -*- coding: utf-8 -*-
# Test detection of the max. possible parallel connection of a winding
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from swat_em.wdggenerator import genwdg
from swat_em.datamodel import datamodel
import swat_em.analyse
import numpy as np
import pdb


def test_parallel_circuit():
    Q         = [12, 12, 9, 12, 27]
    P         = [10,  8, 8,  2, 18]
    m         = [3,   3, 3,  3,  3]
    wstep     = [1,   1, 1,  6,  1]
    layers    = [2,   2, 2,  1,  2]
    a_desired = [2,   4, 1,  2,  9] 

    print('test for max. number of parallel circuits')
    for k in range(len(Q)):
        print('test with Q = {}, 2p = {}, m = {}, wstep = {}, layers = {}'.format(
           Q[k], P[k], m[k], wstep[k], layers[k]))
        ret = genwdg(Q[k], P[k], m[k], wstep[k], layers[k])

        data = datamodel()
        data.set_machinedata(Q = Q[k], p = P[k]/2, m = m[k])
        data.set_phases(ret['phases'])
        bc, txt = data.get_basic_characteristics()
        assert a_desired[k] == bc['a']



if __name__ == '__main__':
    test_parallel_circuit()







# -*- coding: utf-8 -*-

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

#  from swat.wdggenerator import genwdg, overlapping_fractional_slot_slayer, overlapping_fractional_slot_dlayer
from swat_em.wdggenerator import genwdg
from swat_em.datamodel import datamodel
import swat_em.analyse
import numpy as np
import pdb

def test1():
    print('Test double layer toothcoilwinding')
    Q = 18
    p = 8
    m = 3
    wstep = 1
    layers = 2
    #  U = [[1, -2, 3, 10, -11, 12], [-2, 3, -4, -11, 12, -13]]
    #  V = [[4, -5, 6, 13, -14, 15], [-5, 6, -7, -14, 15, -16]]
    #  W = [[7, -8, 9, 16, -17, 18], [-8, 9, -10, -17, 18, -1]]
    U = [[1,  8, -9, 10, 17, -18], [-2, -9, 10, -11, -18, 1]]
    V = [[2, -3, 4, 11, -12, 13], [-3, 4, -5, -12, 13, -14]]
    W = [[5, -6, 7, 14, -15, 16], [-6, 7, -8, -15, 16, -17]]
    kw1 = 0.9452		
    kw5 = -0.1398		
    kw7 = -0.0607
    ret = genwdg(Q, 2*p, m, wstep, layers)
    wdglayout = ret['phases']
    print(wdglayout)
    assert [U, V, W] == wdglayout
    

    data = datamodel()
    #  data.set_config(get_init_config())
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    
    bc, txt = data.get_basic_characteristics()
    print(bc)


def test2():
    print('Test single layer toothcoilwinding')
    Q = 18
    p = 8
    m = 3
    wstep = 1
    layers = 1
    #  U = [[1,-2,3,-4,-11,12]]
    #  V = [[-5,6,13,-14,15,-16]]
    #  W = [[7,-8,9,-10,-17,18]]
    U = [[1, -2, -9, 10, 17, -18],[]]
    V = [[-3, 4, 11, -12, 13, -14],[]]
    W = [[5, -6, 7, -8, -15, 16],[]]
   
    kw1 = 0.9452		
    kw5 = -0.1398		
    kw7 = -0.0607
    ret = genwdg(Q, 2*p, m, wstep, layers) # 12/10
    wdglayout = ret['phases']
    
    print(wdglayout)
    assert [U, V, W] == wdglayout
    
    data = datamodel()
    #  data.set_config(get_init_config())
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    

def test3():
    print('Test single layer toothcoilwinding')
    Q = 12
    p = 4
    m = 3
    wstep = 1
    layers = 1
    U = [[1,-2,7,-8],[]]
    V = [[5,-6,11,-12],[]]
    W = [[3,-4,9,-10],[]]
    kw1 = 0.866
    kw5 = -0.866
    kw7 = 0.866
    ret = genwdg(Q, 2*p, m, wstep, layers) # 12/10
    wdglayout = ret['phases']
    print(wdglayout)
    assert [U, V, W] == wdglayout

    data = datamodel()
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W


def test4():
    print('Test single layer toothcoilwinding')
    Q = 12
    p = 5
    m = 3
    wstep = 1
    layers = 1
    U = [[1,-2,-7,8],[]]
    V = [[-3,4,9,-10],[]]
    W = [[5,-6,-11,12],[]]
    kw1 = 0.9659		
    kw5 = -0.2588
    kw7 = 0.2588
    ret = genwdg(Q, 2*p, m, wstep, layers) # 12/10
    wdglayout = ret['phases']
    print(wdglayout)
    assert [U, V, W] == wdglayout

    data = datamodel()
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W


def test5():
    print('Test single layer toothcoilwinding')
    Q = 24
    p = 11
    m = 3
    wstep = 1
    layers = 1
    U = [[1, -2, -11, 12, -13, 14, 23, -24],[]]
    V = [[-3, 4, -5, 6, 15, -16, 17, -18],[]]
    W = [[7, -8, 9, -10, -19, 20, -21, 22],[]]
    kw1 = 0.9577	
    kw5 = -0.2053
    kw7 = -0.1576
    ret = genwdg(Q, 2*p, m, wstep, layers) # 12/10
    wdglayout = ret['phases']
    print(wdglayout)
    assert [U, V, W] == wdglayout

    data = datamodel()
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W


def test6():
    print('Test single layer full pitch winding')
    Q = 24
    p = 2
    m = 3
    wstep = -1
    layers = 1
    U = [[1, 2, -7, -8, 13, 14, -19, -20],[]]
    V = [[5, 6, -11, -12, 17, 18, -23, -24],[]]
    W = [[-3, -4, 9, 10, -15, -16, 21, 22],[]]
    kw1 = 0.9659	
    kw5 = -0.2588
    kw7 = 0.2588
    #  wdglayout = genwdg(Q, 2*p, m, wstep, layers) # 12/10
    ret = genwdg(Q, 2*p, m, wstep, layers)
    wdglayout = ret['phases']
    print(wdglayout)
    #  pdb.set_trace()
    assert [U, V, W] == wdglayout

    data = datamodel()
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W


def test7():
    print('Test two layer full pitch winding')
    Q = 24
    p = 2
    m = 3
    wstep = 5
    layers = 2
    U = [[1, 2, -7, -8, 13, 14, -19, -20], [-6, -7, 12, 13, -18, -19, 24, 1]]
    V = [[5, 6, -11, -12, 17, 18, -23, -24], [-10, -11, 16, 17, -22, -23, 4, 5]]
    W = [[-3, -4, 9, 10, -15, -16, 21, 22], [8, 9, -14, -15, 20, 21, -2, -3]]
    kw1 = 0.9330	
    kw5 = -0.0670
    kw7 = 0.0670
    ret = genwdg(Q, 2*p, m, wstep, layers) # 12/10
    wdglayout = ret['phases']
    #  wdglayout = overlapping_fractional_slot_dlayer(Q, 2*p, m, wstep)
    print(wdglayout)
    #  pdb.set_trace()
    assert [U, V, W] == wdglayout

    data = datamodel()
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W









"""
# Test is maybe not correct!!!
def test8():
    print('Test overlapping fractional slot winding with single layer')
    # TODO: Referenzdaten ergänzen
    Q = 33
    p = 2
    m = 3
    #  wstep = 1
    layers = 1
    U = [[1, 2, 3, -9, -10, -11, 17, 18, 19, -26, -27], []]
    V = [[6, 7, 8, -15, -16, 23, 24, 25, -31, -32, -33], []]
    W = [[-4, -5, 12, 13, 14, -20, -21, -22, 28, 29, 30], []]
    kw1 = 0.954208
    kw5 = 0.187366
    kw7 = 0.131332
    ret = genwdg(Q, 2*p, m, None, layers) # 12/10
    wdglayout = ret['phases']
    print(wdglayout)
    assert [U, V, W] == wdglayout

    data = datamodel()
    data.set_machinedata(Q = Q, p = p, m = m)
    data.set_phases(wdglayout)
    data.analyse_wdg()
    #  pdb.set_trace()
    idx = data.results['nu_el'].index(1)
    assert kw1 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw1 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw1 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(5)
    assert kw5 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw5 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw5 == np.round(data.results['kw_el'][idx][2], 4) # phase W
    idx = data.results['nu_el'].index(7)
    assert kw7 == np.round(data.results['kw_el'][idx][0], 4) # phase U
    assert kw7 == np.round(data.results['kw_el'][idx][1], 4) # phase V
    assert kw7 == np.round(data.results['kw_el'][idx][2], 4) # phase W
"""



if __name__ == '__main__':
    test1()
    test2()
    test3()
    test4()
    test5()
    test6()
    test7()
    #  test8()








from PyQt5 import uic
from PyQt5.QtWidgets import QDialog
import os
__dir__ = os.path.dirname(os.path.abspath(__file__))



    
class get_notes(QDialog):
    def __init__(self, init_text = ''):
        self.init_text = init_text
        super().__init__()
        uic.loadUi(os.path.join(__dir__, 'ui/Notes.ui'), self)



    def run(self):
        self.plainTextEdit.setPlainText(self.init_text)
        
        ok = self.exec_()
        if ok:
            print(str(self.plainTextEdit.toPlainText()))
            return str(self.plainTextEdit.toPlainText())
        else:
            return None



from PyQt5 import uic
from PyQt5 import QtGui
from PyQt5.QtWidgets import QDialog, QTableWidgetItem, QMessageBox
from swat_em.config import get_phase_color, config
import os
__dir__ = os.path.dirname(os.path.abspath(__file__))



def get_int_from_str(txt):
    '''
    converts txt to a integer number and returns this number
    return None if txt is not a number
    '''
    try:
        number = int(txt)
        if number is not 0:
            return number
    except:
        return None


class layout(QDialog):
    '''
    This is a basic winding editor to define and change the 
    winding layout
    '''
    def __init__(self, data):
        self.data = data
        super().__init__()
        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, 'ui/winding_layout.ui'), self)
        self.setWindowTitle('Edit winding layout')
        
        self.Button_EditMachineData.clicked.connect(self.dialog_machinedata)
        
        #  self.Button_CANCEL.clicked.connect(self.reject)
        #  self.Button_OK.clicked.connect(self.accept)
        
        
        self.tableWindingLayout.cellChanged.connect(self.update_colors)
        self.update_table(self.data.machinedata['phases'])

        self.update_lineEdits(self.data.machinedata['Q'],
                              self.data.machinedata['m'],
                              2*self.data.machinedata['p'],
                              self.data.get_num_layers()
                              )
        
    
    def dialog_machinedata(self):
        '''
        User can change Q, m, P, layers
        '''
        dialog = machinedata(self.data)
        ret = dialog.run()   
        make_update = False
        if ret:
            if ret['Q'] != self.data.machinedata['Q']:
                make_update = True
            if ret['layers'] != self.data.get_num_layers():
                make_update = True
                for km in range(len(self.data.machinedata['phases'])):
                    for ks in range(1, len(self.data.machinedata['phases'][km])):
                        self.data.machinedata['phases'][km][ks] = []
            self.data.set_machinedata(Q = ret['Q'], m = ret['m'], p = ret['P']/2)
            self.update_lineEdits(self.data.machinedata['Q'],
                              self.data.machinedata['m'],
                              2*self.data.machinedata['p'],
                              ret['layers']
                              )
            if make_update:
                self.update_table(self.data.machinedata['phases'], layers=ret['layers'])
            
                
    
        
    def update_lineEdits(self, Q, m, P, layers):
        self.lineEdit_Q.setText(str(Q))
        self.lineEdit_m.setText(str(m))
        self.lineEdit_P.setText(str(P))
        if layers == 1:
            self.radioButton_slayer.setChecked(True)
        else:
            self.radioButton_dlayer.setChecked(True)



    def update_table(self, phases, layers = None):
        self.table = self.tableWindingLayout        
        self.table.setColumnCount(self.data.machinedata['Q'])        
        
        head = ['Layer 1', 'Layer 2']
        if not layers:
            layers = self.data.get_num_layers()
        self.table.setRowCount(layers)
        self.table.setVerticalHeaderLabels(head[:layers])

     
        for km, ph in enumerate(phases):
            col = get_phase_color(km)
            for kl in range(len(ph)):
                layer = ph[kl]
                for cs in layer:
                    if cs > 0:
                        self.table.setItem(kl, cs-1, QTableWidgetItem('+' + str(km+1)))
                    else:
                        self.table.setItem(kl, abs(cs)-1, QTableWidgetItem('-' + str(km+1)))
                    item = self.table.item(kl, abs(cs)-1)
                    if item:
                        item.setBackground(QtGui.QColor(col))

        for k1 in range(self.data.machinedata['Q']):
            self.table.resizeColumnToContents(k1)


        #  self.table.setHorizontalHeaderLabels(['nu'] + self.data.machinedata['phasenames'])
        #  self.table.setVerticalHeaderLabels(['']*np.shape(kw)[0])
        #  afont = QFont()
        #  afont.setBold(True)
        
        #  for k in range(self.data.machinedata['m'] + 1):
            #  self.table.horizontalHeaderItem(k).setFont(afont)
    
    
    
    def read_layout(self):
        '''
        Read layout from table
        '''
        S = []
        for layer in [0, 1]:
            for k in range(self.data.machinedata['Q']):
                item = self.table.item(layer, k)
                if item:
                    txt = str(item.text())
                    num = get_int_from_str(txt)
                    if num:
                        phase = num
                        sign = 1 if phase > 0 else -1
                        while len(S) < abs(phase):
                            S.append([[],[]])
                        S[abs(phase)-1][layer].append( (k+1)*sign )
        return S
        

    def update_colors(self):
        '''
        Update the colors in the table if the user defines phases here
        Also there are some tests for the winding
        '''
        error = []
        warning = []
        
        for layer in [0, 1]:
            for k in range(self.data.machinedata['Q']):
                item = self.table.item(layer, k)
                if item:
                    txt = str(item.text())
                    num = get_int_from_str(txt)
                    if num:
                        phase = abs(num)
                        col = get_phase_color(phase-1)
                        item.setBackground(QtGui.QColor(col))
                    else:
                        item.setBackground(QtGui.QColor('white'))
        
        # Test for errors
        S = self.read_layout()        
        for k in range(len(S)):
            S[k] = [item for sublist in S[k] for item in sublist]  # flatten layers
        
        # test if the number of positve and negative coil sides are equal
        for k in range(len(S)):
            s = S[k]
            pos = 0; neg = 0
            for w in s:
                if w > 0:
                    pos += 1
                elif w < 0:
                    neg += 1
            if pos != neg:
                error.append('Phase {} has {} postive and {} negative coil sides'.format(k+1, pos, neg))

        # test if all phases have the same number of coil sides
        l = [len(s) for s in S]
        if len(set(l)) != 1:
            txt = 'Not all phases have the same number of coil sides:<br>'            
            for k in range(len(S)):
                txt += 'Phase {} hat {} coilsides<br>'.format(k+1, l[k])
            error.append(txt)

        if len(S) > self.data.machinedata['m']:
            error.append('There are {} phases but in main window only m = {} is set'.format(
                len(S), self.data.machinedata['m']))

        # print errors
        txt_error = '<span style=\" color:#ff0000;\" >'
        for e in error:
            txt_error += str(e) + '<br>'
        txt_error += '</span>'
        self.textBrowser_output.setHtml(txt_error)


        
    def run(self):
        ok = self.exec_()
        if ok:
            if self.radioButton_overwrite_winding.isChecked():
                overwrite = True
            else:
                overwrite = False
            phases = self.read_layout()
            layer1 = False; layer2 = False  # test if single or double layer winding
            for s in phases:
                if len(s[0]) > 0:
                    layer1 = True
                if len(s[1]) > 0:
                    layer2 = True
            if layer1 and layer2:
                layers = 2
            else:
                layers = 1
            
            Q = self.data.machinedata['Q']
            p = self.data.machinedata['p']
            m = self.data.machinedata['m']
            w = self.data.machinedata['wstep']       
            ret = {'phases': phases, 'Q': Q, 'P': 2*p, 'm': m, 'w': w, 
            'layers': layers, 'overwrite': overwrite}
            return ret
        else:
            return None
        

        
        
        

        
    



class machinedata(QDialog):
    '''
    User can change Q, m, P, layers
    '''
    def __init__(self, data):
        self.data = data
        super().__init__()
        uic.loadUi(os.path.join(__dir__, 'ui/MachineData.ui'), self)
        self.setWindowTitle('Change machine data')
        self.data = data
        self.spinBox_Q.setValue(self.data.machinedata['Q'])
        self.spinBox_m.setValue(self.data.machinedata['m'])
        self.spinBox_P.setValue(2*self.data.machinedata['p'])
        if self.data.get_num_layers() == 1:
            self.radioButton_slayer.setChecked(True)
        else:
            self.radioButton_dlayer.setChecked(True)
        

    def run(self):
        ok = self.exec_()
        if ok:
            if self.radioButton_slayer.isChecked():
                layers = 1
            elif self.radioButton_dlayer.isChecked():
                layers = 2
            Q = self.spinBox_Q.value()
            P = self.spinBox_P.value()
            m = self.spinBox_m.value()
            if Q < self.data.machinedata['Q']:
                buttonReply = QMessageBox.question(self, 'Warning',
                 '''The defined number of slots is lower than the number of slots in the actual winding. \
Is it ok to loose some these informations of the actual winding layout?''', 
                QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
                if buttonReply != QMessageBox.Yes:
                    return None
            
            if layers < self.data.get_num_layers():
                buttonReply = QMessageBox.question(self, 'Warning',
                 '''The defined number of layers is lower than the number of layers in the actual winding. \
Is it ok to loose some these informations of the actual winding layout?''', 
                QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
                if buttonReply != QMessageBox.Yes:
                    return None
            
            ret = {'Q': Q, 'P': P, 'm': m, 'layers': layers}
            return ret
        else:
            return None












# -*- coding: utf-8 -*-

from PyQt5 import uic
from PyQt5.QtWidgets import QDialog, QApplication, QMessageBox, QTableWidgetItem
from PyQt5.QtGui import QIntValidator
from PyQt5 import QtGui
from PyQt5 import QtCore
#  from PyQt5.QtCore import Qt
import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from swat_em import wdggenerator
from swat_em import datamodel
from swat_em.config import config, get_phase_color

__dir__ = os.path.dirname(os.path.abspath(__file__))


class NewWinding(QDialog):
    def __init__(self):
        super().__init__()
        uic.loadUi(os.path.join(__dir__, 'ui/choose_generator.ui'), self)
        self.generator = None

        self.Button_EditWindingLayout.clicked.connect(lambda: self.ret(generator = 0))
        self.Button_GenerateAutomatically.clicked.connect(lambda: self.ret(generator = 1))
        self.Button_FindByTable.clicked.connect(lambda: self.ret(generator = 2))
    
    def run(self):
        ok = self.exec_()
        if ok:
            return self.generator
        else:
            return None
    
    def ret(self, generator):
        '''saves the choosen generator and say "ok" to the dialog'''
        self.generator = generator
        self.accept()

        



class GenWinding2(QDialog):
    def __init__(self):
        super().__init__()
       
        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, 'ui/GenWinding2.ui'), self)
        self.setWindowTitle('Generate winding')

        self.radioButton_slayer.toggled.connect(self.change_layers) # disable lineEdit for single phase windings
        
        # update 'w' and update winding
        self.spinBox_Q.valueChanged.connect(self.QmP_changed)
        self.spinBox_P.valueChanged.connect(self.QmP_changed)
        self.spinBox_m.valueChanged.connect(self.QmP_changed)
        
        # update winding
        self.radioButton_slayer.toggled.connect(self.update)
        self.spinBox_w.valueChanged.connect(self.update)
        
        self.update()  # calc initial winding
        
        
    def change_layers(self):
        '''
        disable lineEdit for single phase windings
        '''
        if self.radioButton_slayer.isChecked():
            self.spinBox_w.setEnabled(False)
        else:
            self.spinBox_w.setEnabled(True)
        
    def QmP_changed(self):
        Q = self.spinBox_Q.value()
        P = self.spinBox_P.value()
        m = self.spinBox_m.value()
        w = Q//P    # take smaller value (integer div.)
        if w <= 0:
            w = 1
        self.spinBox_w.setValue(w)
        self.update()
        
    def update(self):
        self.Q = self.spinBox_Q.value()
        self.P = self.spinBox_P.value()
        self.m = self.spinBox_m.value()
        self.w = self.spinBox_w.value()
        
        if self.radioButton_slayer.isChecked():
            self.layers = 1
        elif self.radioButton_dlayer.isChecked():
            self.layers = 2

        self.data = datamodel.datamodel()
        self.data.set_machinedata(Q = self.Q, m = self.m, p = self.P/2)
        ret = wdggenerator.genwdg(self.Q, self.P, self.m, self.w, self.layers)
        self.data.set_phases(S = ret['phases'], wstep = ret['wstep'])
        self.data.set_valid(ret['valid'], ret['error'])
        bc, bc_str = self.data.get_basic_characteristics()
        #  self.textBrowser_wdginfo.setPlainText(bc_str)
        self.textBrowser_wdginfo.setHtml(bc_str)
        
        
        self.table = self.tableWindingLayout
        self.table.clear()
        if bc['sym']:            
            self.table.setRowCount(self.layers)
            self.table.setColumnCount(self.data.machinedata['Q'])
            
            for km, ph in enumerate(self.data.machinedata['phases']):
                col = get_phase_color(km)
                for kl in range(len(ph)):
                    layer = ph[kl]
                    for cs in layer:
                        if cs > 0:
                            self.table.setItem(kl, cs-1, QTableWidgetItem('+' + str(km+1)))
                        else:
                            self.table.setItem(kl, abs(cs)-1, QTableWidgetItem('-' + str(km+1)))
                        self.table.item(kl, abs(cs)-1).setBackground(QtGui.QColor(col))
                        #  self.table.item(kl, abs(cs)-1).setFlags(Qt.ItemIsEditable)
                        self.table.item(kl, abs(cs)-1).setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
            for k1 in range(self.data.machinedata['Q']):
                #  self.table.resizeColumnToContents(k1)
                self.table.setColumnWidth(k1, 25)


    def run(self):
        ok = self.exec_()
        if ok:
            if self.radioButton_overwrite_winding.isChecked():
                overwrite = True
            else:
                overwrite = False
            ret = {'Q': self.Q, 'P': self.P, 'm': self.m, 'w': self.w, 
            'layers': self.layers, 'overwrite': overwrite}
            return ret
        else:
            return None




class GenWindingCombinations(QDialog):
    def __init__(self):
        super().__init__()
       
        uic.loadUi(os.path.join(__dir__, 'ui/GenCombinations.ui'), self)
        self.setWindowTitle('Generate windings per table')
       
        self.colorlabel1.setStyleSheet("background-color:#ADD8E6;");
        self.colorlabel2.setStyleSheet("background-color:#E6C3AD;");
        self.colorlabel3.setStyleSheet("background-color:#BCFFBC;");
        
        # update 'w' and update winding
        self.spinBox_Q1.valueChanged.connect(self.generate)
        self.spinBox_Q2.valueChanged.connect(self.generate)
        self.spinBox_P1.valueChanged.connect(self.generate)
        self.spinBox_P2.valueChanged.connect(self.generate)
        self.spinBox_m.valueChanged.connect(self.generate)
        self.radioButton_slayer.toggled.connect(self.generate)
        self.checkBox_toothcoil.toggled.connect(self.generate)
        
        self.tableCombinations.itemSelectionChanged.connect(self.combination_selected)
        
        # Fill combobox
        self.comboBox_plotval.addItems(['kw1 (winding factor)',
                                        'q (slots per pole per phase)',
                                        't (symmetry)',
                                        'a (parallel circuit)',
                                        'lcm(Q,P) (least common multiple)'])
        self.comboBox_plotval.currentIndexChanged.connect(self.update_table)
        #  self.generate()

        
    def generate(self):
        '''
        Generate possible winding layout for different slot/pole combinations
        '''
        self.Q1 = self.spinBox_Q1.value()
        self.Q2 = self.spinBox_Q2.value()
        self.P1 = self.spinBox_P1.value()
        self.P2 = self.spinBox_P2.value()
        self.m = self.spinBox_m.value()
        if self.checkBox_toothcoil.isChecked():
            wstep = 1
        else:
            wstep = -1
        
        if self.radioButton_slayer.isChecked():
            self.layers = 1
        elif self.radioButton_dlayer.isChecked():
            self.layers = 2

        self.Qlist = []
        for k in range(self.Q1, self.Q2+1, 1):
            if k % self.m == 0:
                self.Qlist.append(k)
        #  self.Qlist = list(range(self.Q1, self.Q2+1, self.m))
        self.Plist = list(range(self.P1, self.P2+1, 2))
        
        self.data = []
        for iQ, kQ in enumerate(self.Qlist):
            self.data.append([])
            for iP, kP in enumerate(self.Plist):
                d = datamodel.datamodel()
                d.set_machinedata(Q = kQ, m = self.m, p = kP/2)
                ret = wdggenerator.genwdg(kQ, kP, self.m, wstep, self.layers)
                d.set_phases(S = ret['phases'], wstep = ret['wstep'])
                d.set_valid(ret['valid'], ret['error'])
                bc, bc_str = d.get_basic_characteristics()
                self.data[iQ].append(d)
        #  print(self.data)
        #  import pdb
        #  pdb.set_trace()
        self.update_table()
                
        
    def update_table(self):        
        self.table = self.tableCombinations
        self.table.clear()
       
        self.table.setRowCount(len(self.Qlist))
        self.table.setColumnCount(len(self.Plist))
        
        for i in range(len(self.Plist)):
            table_header_item = QTableWidgetItem('2p='+str(self.Plist[i]))
            myFont=QtGui.QFont()
            myFont.setBold(True)
            table_header_item.setFont(myFont)
            self.table.setHorizontalHeaderItem(i, table_header_item)

        for i in range(len(self.Qlist)):
            table_header_item = QTableWidgetItem('Q='+str(self.Qlist[i]))
            myFont=QtGui.QFont()
            myFont.setBold(True)
            table_header_item.setFont(myFont)
            self.table.setVerticalHeaderItem(i, table_header_item)        
        
        for iQ, kQ in enumerate(self.Qlist):
            self.data.append([])
            for iP, kP in enumerate(self.Plist):
                if self.data[iQ][iP].results['valid'] and self.data[iQ][iP].results['basic_char']['sym']:
                    bc = self.data[iQ][iP].results['basic_char']
                    if bc['kw1'][0] > 0.01:
                        idx = self.comboBox_plotval.currentIndex()
                        if idx == 0: 
                            txt = str(round(bc['kw1'][0], 3))
                        elif idx == 1:
                            txt = str(bc['q'])
                        elif idx == 2:
                            txt = str(bc['t'])
                        elif idx == 3:
                            txt = str(bc['a'])
                        elif idx == 4:
                            txt = str(bc['lcmQP'])

                        self.table.setItem(iQ, iP, QTableWidgetItem(txt))
                        q = bc['q']

                        if self.data[iQ][iP].machinedata['wstep'] == 1: 
                            self.table.item(iQ, iP).setBackground(QtGui.QColor('#ADD8E6'))
                        elif q.denominator == 1:
                            self.table.item(iQ, iP).setBackground(QtGui.QColor('#E6C3AD'))
                        else:
                            self.table.item(iQ, iP).setBackground(QtGui.QColor('#BCFFBC'))
                        self.table.item(iQ, iP).setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled) # note editable

        for iP, kP in enumerate(self.Plist):
            self.table.resizeColumnToContents(iP)
            #  self.table.setColumnWidth(k1, 25)
        
        # reset table with winding layout because there is noting choosed
        self.tableWindingLayout.clear()
        self.tableWindingLayout.setRowCount(0)
        self.tableWindingLayout.setColumnCount(0)  


    def combination_selected(self):
        '''
        user selected a cell/winding combination
        '''
        sel = self.tableCombinations.selectedIndexes()
        if sel:
            sel = sel[0]        
            row = sel.row()
            column = sel.column()
            d = self.data[row][column]
            bc, bc_text = d.get_basic_characteristics()
            self.textBrowser_wdginfo.setHtml(bc_text)
            
            self.table = self.tableWindingLayout
            self.table.clear()
            self.table.setRowCount(self.layers)
            self.table.setColumnCount(d.machinedata['Q'])            
            for km, ph in enumerate(d.machinedata['phases']):
                col = get_phase_color(km)
                for kl in range(len(ph)):
                    layer = ph[kl]
                    for cs in layer:
                        if cs > 0:
                            self.table.setItem(kl, cs-1, QTableWidgetItem('+' + str(km+1)))
                        else:
                            self.table.setItem(kl, abs(cs)-1, QTableWidgetItem('-' + str(km+1)))
                        if abs(cs) <= d.machinedata['Q']: # for non-valid windings there could be a slot-number that is bigger than the number of slots
                            self.table.item(kl, abs(cs)-1).setBackground(QtGui.QColor(col))
                        self.table.item(kl, abs(cs)-1).setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled) # note editable
                        
            for k1 in range(d.machinedata['Q']):
                self.table.resizeColumnToContents(k1)
            return row, column        
        else:
            return None, None
        
  
    def run(self):
        self.generate()
        ok = self.exec_()
        if ok:
            if self.radioButton_overwrite_winding.isChecked():
                overwrite = True
            else:
                overwrite = False
            row, column = self.combination_selected()
            if row is not None:
                ret = {}
                ret['Q'] = self.data[row][column].machinedata['Q']
                ret['P'] = 2*self.data[row][column].machinedata['p']
                ret['m'] = self.data[row][column].machinedata['m']
                ret['w'] = self.data[row][column].machinedata['wstep']
                ret['layers'] = self.layers
                ret['overwrite'] = overwrite
                return ret
            else:
                return None
        else:
            return None














# -*- coding: utf-8 -*-
'''
Provides functions for analyzing windings
''' 
import numpy as np
import fractions
import math
from collections import Counter


def calc_q(Q, p, m):
    return fractions.Fraction(Q / (m * 2*p)).limit_denominator(100)

def get_basic_characteristics(Q, P, m, S, turns=1):
    q = fractions.Fraction(Q / (m * P)).limit_denominator(100)
    Ei, kw = calc_star(Q, S, turns, P/2, 1)
    a = wdg_get_periodic([Ei], S)
    a = a[0]  # phase 1
    sym = wdg_is_symmetric([Ei], m)
    t = math.gcd(int(Q), int(P/2))
    lcmQP = int(np.lcm(int(Q), int(P)))
    bc = {'q': q, 'kw1': kw, 'a': a, 'sym': sym, 't': t, 'lcmQP': lcmQP}
    return bc
    

def wdg_get_periodic(Ei, S):
    '''
    Returns the symmetry factor for the winding 

    Parameters
    ----------
    Ei :     list of lists of lists
             voltage vectors for every phase and every slot
             Ei[nu][phase][slot]
    S :      list of lists
             winding layout
             
    Returns
    -------
    return : list
             symmetry factor for each phase. 
             1 if there is no periodicity
             2 if half of the machine is smallest symmetric part and so on...
    ''' 
    if len(Ei) == 0:
        return 1
    if len(Ei[0]) == 0:
        return 1
    
    periodic = []
    # for each phase
    for km in range(len(S)):
        ei = Ei[0][km] # only for fundamental
        S2 = S[km]
        S2 = [item for sublist in S2 for item in sublist]  # flatten layers
        ei_pos = []; ei_neg = []
        for i, s in enumerate(S2):
            if s>0:
                ei_pos.append(ei[i])  # phasors of pos. coil sides
            else:
                ei_neg.append(ei[i])
        if len(ei_pos) != len(ei_neg):
            periodic.append(1)
        else:
            a_max = 0
            #  a_list = [1]
            for k in range(len(ei_pos)):   # all combinations of connections of coil sides
                angles = np.angle(ei_pos + np.roll(ei_neg, k))
                angles = np.round(angles, 4)
                c = Counter(angles)
                a = min(c.values())
                if a > a_max:
                    a_max = a
            periodic.append( a_max )
    return periodic


def wdg_is_symmetric(Ei, m):
    '''
    Test if the winding ist symmetric -> phase shift between every 
    phase is 360°/m. 

    Parameters
    ----------
    Ei :     list of lists of lists
             voltage vectors for every phase and every slot
             Ei[nu][phase][slot]
    m  :     integer
             number of phases     
             
    Returns
    -------
    return : bool
             True if symmetric, False if not

    ''' 
    if len(Ei) == 0:
        return None
    if len(Ei[0]) == 0:
        return None
    ei = Ei[0] # only for fundamental
    angles = [np.angle(np.sum(k))/np.pi*180 for k in ei]
    angles = [np.round(k, 4) for k in angles]
    length = [np.abs(np.sum(k)) for k in ei]
    #  print('angles1', angles)
    if m % 2 == 0: # shift all phasors to >180 for even number of phases
        angles = np.array(angles)
        angles += 360
        angles -= min(angles)
        for k in range(len(angles)):
            if angles[k] < 180:
                angles[k] += 360
        angles = angles.tolist()

    angles.sort()
    #  print('angles2', angles)
    if m%2 == 0:
        target_angle = 360 / (2*m)
    else:
        target_angle = 360 / m
    diff = np.diff(angles)
    sym = True
    #  print('diff', diff)
    # Test for phase angles
    for d in diff:
        if not np.isclose(target_angle, d, rtol=1e-02, atol=1e-02):
            sym = False
    # Test for amplitude
    for k in range(1, len(length)):
        if not np.isclose(length[0], length[k], rtol=1e-02, atol=1e-02):
            sym = False
    return sym


def calc_phaseangle_starvoltage(Ei):
    '''
    Calculates the phaseangle based on the slot voltage vectors. 

    Parameters
    ----------
    Ei :     list of lists of lists
             voltage vectors for every phase and every slot
             Ei[nu][phase][slot]
             
    Returns
    -------
    return phaseangle: list
                       phaseangle for every harmonic number and every
                       phase; phaseangle[nu][phase]
    return seqeunce:   list
                       sequence of the flux wave: 1 or -1
    ''' 
    phaseangle = []
    for knu in Ei:             # for every nu
        phaseangle.append([])
        angle = [np.angle(sum(km))*180/np.pi for km in knu] # for every phase
        for a in angle:
            while a < 0:
                a += 360.0
            phaseangle[-1].append( a ) 

    sequence = []
    for p in phaseangle:
        if p[1] > p[0]:
            sequence.append(1)
        else:
            sequence.append(-1)
    #  if sequence[0] < 0:  # related to the fundamental
        #  sequence = [-k for k in sequence]

    return phaseangle, sequence


def calc_kw(Q, S, turns, p, N_nu, config):
    '''
    Calculates the windingfactor, the slot voltage vectors. The 
    harmonic numbers are generated automatically.

    Parameters
    ----------
    Q :      integer
             number of slots
    S :      list of lists
             winding layout
    turns :  number or list of lists (shape of 'S')
             number of turns. If turns is a list of lists, for each
             coil side a specific number of turns is used
    p :      integer
             number of pole pairs
    N_nu:    integer
             length of the harmonic number vector
             
    Returns
    -------
    return nu: list
               harmonic numbers with relevant winding factor
    return Ei: list of lists of lists
               voltage vectors for every phase and every slot
               Ei[nu][phase][slot]
    return wf: list of lists
               winding factor for every phase. The sign defines the
               direction of the flux wave
               wf[nu][phase]
    '''
    #  S = [item for sublist in S for item in sublist]  # flatten layers
    nu = [] # order
    wf = [] # winding factor
    Ei = [] # slot voltage vectors
    k = 1
    while len(wf) < N_nu:
        a, b = calc_star(Q, S, turns, p, k)
        if np.all(np.abs(b) > config['kw_min']):
            nu.append(k)
            wf.append(b)
            Ei.append(a)
        k+=1
        if k > 10000:   # break infinity loop if there is no relevant 
            break       # windingfactor of the actual winding layout 

    phase, sequence = calc_phaseangle_starvoltage(Ei)
    for k in range(len(sequence)):
        wf[k] = [sequence[k] * s for s in wf[k]]
    
    return nu, Ei, wf, phase
    
   

def calc_star(Q, S, turns, p, nu):
    '''
    Calculates the slot voltage vectors for the given winding layout 

    Parameters
    ----------
    Q :      integer
             number of slots
    S :      list of lists
             winding layout
    turns :  number or list of lists (shape of 'S')
             number of turns. If turns is a list of lists, for each
             coil side a specific number of turns is used
    p :      integer
             number of pole pairs
    nu:      integer
             harmonic number for calculation
             
    Returns
    -------
    return Ei: list
               voltage vectors for every phase and every slot
               Ei[phase][slot]
    return kw: list
               winding factor (absolute value) for every phase
               kw[phase]
    '''
    S2 = []
    turns2 = []
    for k in range(len(S)):
        S2.append( [item for sublist in S[k] for item in sublist] )  # flatten layers
        #  S2[k] = sorted(S2[k], key=abs)  # sort to abs values (slots in sequence -1,2,2,-3,...
        idx = np.argsort(np.abs(S2[k]))
        S2[k] = np.array(S2[k])[idx]
        if hasattr(turns, '__iter__'):
            turns2.append( [item for sublist in turns[k] for item in sublist] )  # flatten
            turns2[k] =  np.array(turns2[k])[idx]

    Ei = []
    kw = []
    for i, s in enumerate(S2):
        if hasattr(turns, '__iter__'):
            turn = turns2[i]
        else:
            turn = turns
        
        alpha = 2.*nu*p*np.pi/(Q) * np.abs(s)
        
        for k in range(len(s)):
            if s[k] < 0:
                alpha[k] += np.pi           # negative Zeiger umklappen
        
        a = ( turn*np.exp(1j*alpha) )            # Spannungsvektoren berechnen
        if np.sum(a) != 0.0:
            b = np.abs(np.sum(a)) / np.sum(np.abs(a))
        else:
            b = np.array(0.0)
        Ei.append( a.tolist() )
        kw.append( b.tolist() )
    return Ei, kw



def calc_MMK(Q, m, S, turns = 1, N = 3601, angle = 0):
    '''
    Calculates the magneto-motoric force (MMK) 

    Parameters
    ----------
    Q :      integer
             number of slots
    m :      integer
             number of phases
    S :      list of lists
             winding layout
    turns :  number or list of lists (shape of 'S')
             number of turns. If turns is a list of lists, for each
             coil side a specific number of turns is used
    N :      integer
             number of values for the MMK curve
    angle:   float
             actual phase of the current system in deg
             
    Returns
    -------
    return MMK:   list
                  MMK curve
    return theta: list
                  effective current for each slot
    '''
    S2 = []
    turns2 = []
    for k in range(len(S)):
        S2.append( [item for sublist in S[k] for item in sublist] )  # flatten layers
        #  S2[k] = sorted(S2[k], key=abs)  # sort to abs values (slots in sequence -1,2,2,-3,...
        idx = np.argsort(np.abs(S2[k]))
        S2[k] = np.array(S2[k])[idx]
        if hasattr(turns, '__iter__'):
            turns2.append( [item for sublist in turns[k] for item in sublist] )  # flatten
            turns2[k] =  np.array(turns2[k])[idx]
        
    # step function
    def h(x):
        return np.where(x<0., 0., 1.)

    phi = np.linspace(0, 2*np.pi, N)
    I = []
    km = 2 if m%2 == 0 else 1
    for k in range(m):
        I.append( np.cos( 2*np.pi/(m*km)*k - angle/180*np.pi ) )
    theta = np.zeros(Q)
    for k1 in range(m):
        phase = S2[k1]
        
        for k2 in range(len(phase)):
            if hasattr(turns, '__iter__'):
                turn = turns2[k1][k2]
            else:
                turn = turns
            
            idx = np.abs(phase[k2])
            VZ = np.sign(phase[k2])
            theta[idx-1] += VZ*I[k1]*turn
    MMK = np.zeros(np.shape(phi))
    for k in range(Q):
        MMK += theta[k] * h(phi-2*np.pi/Q*k)
    MMK -= np.mean(MMK[:-1])
    phi = phi/max(phi)*Q
    return phi.tolist(), MMK.tolist(), theta.tolist()



def DFT(vect):
    """
    Harmonic Analyses
    
    Parameters
    ----------
    vect  : array_like
            curve (time signal)
             
    Returns
    -------
    return : complex ndarray
             complex spectrum
    """
    #  vect = np.array(vect,float)
    N = len(vect)
    yy = 2./N*np.fft.fft(vect)
    yy[0] = np.mean(vect)
    return yy  



